#include "snn_gold.h"

#define M    100
#define N    1000
void main()
{
	// create a network with N neurons, maximum number of synapes is 100, maximum delay is 20ms, name is global
	CpuSNN s(N, M, 20, "global");

	// create a population or group of neurons with the label "excit", having 80,000 neurons, and respective izhikevich parameters (a,b,c,d)
	// another function exists where you can specify for each of the Izhikevich parameters as (mean, std. deviation) pair. This allows random distribution for a,b,c,d parameters
	int g1=s.createInitGroup("excit", N*0.8, 0.02f, 0.2f, -65.0f, 8.0f);

	// create a population or group of neuron with the label "inhib", having 20,000 neurons and respective Izhikevich parameters(a,b,c,d)
	int g2=s.createInitGroup("inhib", N*0.2, 0.1f,  0.2f, -65.0f, 2.0f);

	// connect the group g2 (inhib) to the group g1(excit). The connection is random and have M synapses from each neuron in g2 to g1 with initial weight of -5.0. The maximum and minimum delay is 1ms
	s.connect(g2,g1,"random", -5.0f, M, 1, 1);

	// connect the group g1(excit) to the group g2(inhib). The connection is of random type and have 0.2*M synapses for each neuron in g1 to g2 with initial weight of +6.0. Delay range is 1ms and 20ms.
	s.connect(g1,g2,"random", +6.0f, M*0.2,  1, 20);

	// connect the group g1(excit) to the group g1(excit). This is a self connection. Make 0.8*M synapses randomly to the neurons in g1 with initial weight of +6.0. The delay range is 1ms and 20ms.
	s.connect(g1,g1,"random", +6.0f, M*0.8,  1, 20);

	// here I define the property of the weight.
	s.snnProperty("smax",10.0);

	// here I define the property of the thalamic current into the network. select 0.1% of neurons ( 0.001*N neurons) and apply a thalamic current with the strength of 20.0
	s.randomNoiseCurrent(0.1, 20.0);

	// here we define and set the properties of the STDP. 
	float ALPHA_LTP = 0.10f;
	float TAU_LTP   = 20.0f;
	float ALPHA_LTD = 0.12f;
	float TAU_LTD   = 20.0f;	
	s.stdp( ALPHA_LTP, TAU_LTP, ALPHA_LTD, TAU_LTD);

	// show logout every 1 sec, if value is set to 0..disable logging. You can pass a file pointer or pass stdout to specify where the log output should go.
	s.setLogCycle(1, stdout);		

	// here we create a spike monitor. This enables loading of the generated spikes for specific population of neurons in group g1
	s.spikeMonitor(g1);

	// here we create another spike monitor. This allows dumping the firing properties of a population of neurons in a group g2
	s.spikeMonitor(g2);

	for(int i=0; i < 25; i++) {

		// run the established network for a duration of 1 (sec)  and 0 (millisecond), in GPU_MODE
		s.runNetwork(1,0, GPU_MODE);

		// after running the simulation, dump the synaptic weight histogram in a matlab file with the given name (synhist) and in directory "Histogram". The given file name can have directory structure in it.
		// If the second argument is TRUE, then dont overwrite the created histogram file. One new file is created for each call with given filename and additional postfix.
		s.printHistogram("Histogram/synhist", true);		

		// Dump the firing information in a file Raster/fireInfo1.m. If second argument is true, dont overwrite the file but create a new file with same name and different postfix
		s.rasterPlot(g1, "Raster/fireInfo1.m", true);	

		s.rasterPlot(g2, "Raster/fireInfo2.m", true);
	}

	// display the details of the current simulation run
	s.printSimSummary();

}

/*
#define NUMBER_SYNAPSES 100
#define NUMBER_NEURONS	100000

int main()
{

	CpuSNN s(NUMBER_NEURONS, NUMBER_SYNAPSES, 20, "global");

	int g1=s.createInitGroup("excit", NUMBER_NEURONS*0.8, 0.02f, 0.2f, -65.0f, 8.0f);
	int g2=s.createInitGroup("inhib", NUMBER_NEURONS*0.2, 0.1f,  0.2f, -65.0f, 2.0f);

	#ifdef CONDUCTANCE
	s.connect(g2,g1,"random", -0.2f, NUMBER_SYNAPSES, 1, 1);
	s.connect(g1,g2,"random", +0.2f, NUMBER_SYNAPSES*0.2,  1, 5);
	s.connect(g1,g1,"random", +0.2f, NUMBER_SYNAPSES*0.8,  1, 5);

	s.snnProperty("smax",0.5);	

	s.randomNoiseCurrent(1.0, 0.0);
	//s.randomNoiseCurrent(1.0, 20.0, g2);
	#else

	s.connect(g2,g1,"random", -5.0f, NUMBER_SYNAPSES, 1, 1);
	s.connect(g1,g2,"random", +6.0f, NUMBER_SYNAPSES*0.2,  1, 20);
	s.connect(g1,g1,"random", +6.0f, NUMBER_SYNAPSES*0.8,  1, 20);

	s.snnProperty("smax",10.0);	

	s.randomNoiseCurrent(0.1, 20.0);
	//s.randomNoiseCurrent(0.1, 30.0, g2);

	#endif

	float ALPHA_LTP = 0.10f;
	float TAU_LTP   = 20.0f;
	float ALPHA_LTD = 0.12f;
	float TAU_LTD   = 20.0f;
	
	s.stdp( ALPHA_LTP, TAU_LTP, ALPHA_LTD, TAU_LTD);

	s.setLogCycle(1, stdout);		// show logout every 1 sec, if set to 0..disable logging..

	//s.spikeMonitor(g1);

	//s.spikeMonitor(g2);

	//s.printConnection("net.log");
	
	char fname[100];

    
	for(int i=0; i < 25; i++) {

		s.runNetwork(1,0, GPU_MODE);

		//s.printHistogram("Histogram/synhist", true);		

		//s.rasterPlot(g1, "Raster/fireInfo1.m", true);	

		//s.rasterPlot(g2, "Raster/fireInfo2.m", true);
	}

	//save state from GPU to CPU
	s.saveStateGPUtoCPU();

	//run network for 10sec and measure timing in GPU_MODE
	s.startGPUTiming();
	s.runNetwork(10,0,GPU_MODE);
	s.stopGPUTiming();
		
	//run network for 10sec and measure timing in CPU_MODE
	s.startCPUTiming();
	s.runNetwork(10,0,CPU_MODE);
	s.stopCPUTiming();
	s.printSimSummary();
}
*/