#include "snn_gold.h"

#define NUMBER_SYNAPSES 100
#define NUMBER_NEURONS	1000

int main()
{
	CpuSNN s(NUMBER_NEURONS, NUMBER_SYNAPSES, 20, "global");

	int g1=s.createInitGroup("excit", NUMBER_NEURONS*0.8, 0.02f, 0.2f, -65.0f, 8.0f);
	int g2=s.createInitGroup("inhib", NUMBER_NEURONS*0.2, 0.1f,  0.2f, -65.0f, 2.0f);

	#ifdef CONDUCTANCE
	s.connect(g2,g1,"random", -0.2f, NUMBER_SYNAPSES, 1, 1);
	s.connect(g1,g2,"random", +0.2f, NUMBER_SYNAPSES*0.2,  1, 5);
	s.connect(g1,g1,"random", +0.2f, NUMBER_SYNAPSES*0.8,  1, 5);

	s.snnProperty("smax",0.5);	

	s.randomNoiseCurrent(1.0, 0.0);
	//s.randomNoiseCurrent(1.0, 20.0, g2);
	#else
	s.connect(g2,g1,"random", -5.0f, NUMBER_SYNAPSES, 1, 1);
	s.connect(g1,g2,"random", +4.0f, NUMBER_SYNAPSES*0.2,  1, 5);
	s.connect(g1,g1,"random", +4.0f, NUMBER_SYNAPSES*0.8,  1, 5);

	s.snnProperty("smax",10.0);	

	s.randomNoiseCurrent(1.0, 30.0, g1);
	s.randomNoiseCurrent(1.0, 30.0, g2);
	#endif

	float ALPHA_LTP = 0.10f;
	float TAU_LTP   = 20.0f;
	float ALPHA_LTD = 0.12f;
	float TAU_LTD   = 20.0f;
	
	s.stdp( ALPHA_LTP, TAU_LTP, ALPHA_LTD, TAU_LTD);

	s.setLogCycle(10, stdout);		// show logout every 1 sec, if set to 0..disable logging..

	s.spikeMonitor(g1);

	s.spikeMonitor(g2);

	s.printConnection("net.log");
	
	char fname[100];

	for(int i=0; i < 10; i++) {

		s.runNetwork(1,0, GPU_MODE);
	
		s.printHistogram("Histogram/synhist", true);		

		s.rasterPlot(g1, "Raster/fireInfo1.m", true);	

		s.rasterPlot(g2, "Raster/fireInfo2.m", true);

	}

/*
	/*
	//save state from GPU to CPU
	s.saveStateGPUtoCPU();

	//run network for 10sec and measure timing in GPU_MODE
	s.startGPUTiming();
	s.runNetwork(10,0,GPU_MODE);
	s.stopGPUTiming();
		
	//run network for 10sec and measure timing in CPU_MODE
	s.startCPUTiming();
	s.runNetwork(10,0,CPU_MODE);
	s.stopCPUTiming();
	*/
	s.printSimSummary();
}
