#ifndef _SNN_GOLD_H_
#define _SNN_GOLD_H_

#include "mtrand.h"
#include "gpu_random.hpp"
#include "config.h"

#include <cutil_inline.h>
#include <cuda.h>

//Used by all network to generate global random number...
extern RNG_rand48* gpuRand48;

typedef struct noiseGenProperty_s
{
	float		neuronPercentage;	// percentage of neuron that needs random inputs..
	float		currentStrength;	// strength of the noise current applied...
	int			groupId;			// group id.. cross references to group properties
	int			nstart;				// start of the neuron id
	int			ncount;				// total neuron in current group
	int			rand_ncount;		// (ncount*neuronPercentage/100) neurons will be selected
} noiseGenProperty_t;

typedef struct fileInfo_s
{
	FILE* fp;
	char* fname;
	struct fileInfo_s* next;
} fileInfo_t;

typedef struct {
    short  delay_index_start;
    short  delay_length;
} delay_info_t;

typedef struct network_info_s
{
	int			N,M,D,Ne,Ni;
	int			I_setLength;
	size_t		I_setPitch;
	int			preSynLength;	
	float		ALPHA_LTP, TAU_LTP;
	float		ALPHA_LTD, TAU_LTD;
	bool		STDPenabled;
	int			numRandNeurons;
	int			numNoise;
	int			postSynCnt;
	int			maxSpikesExcitatory,maxSpikesInhibitory;
	float		maxWeight;	// maximum value for the synaptic weights
	float		maxWeightChange;  // maximum value reached by sd
} network_info_t;

typedef struct network_ptr_s
{
	int	memType;
	int	allocated;		// true if all data has been allocated..
	
	float	*v, *u, *a, *b, *c, *d, *I;		
	short	*Npre;			// stores the number of input connections to the neuron			
	short	*Npre_exc;		// stores the number of excitatory input connection to the input
	short	*Npost;			// stores the number of output connections from a neuron.
	int	*lastSpikeTime;	// storees the firing time of the neuron
	float	*wtChange, *wt;		// stores the synaptic weight and weight change of a synaptic connection
	int	*synSpikeTime;
	int*	I_set;
	int	*cumulativePost;
	int	*cumulativePre;
	int	*postSynapticIds;		// 10 bit syn id, 22 bit neuron id, ordered based on delay
								// allows maximum synapses of 1024 and maximum network
								// size of 4 million neurons, with 64 bit representation
								// we can have larger networks for simulation
	delay_info_t    *postDelayInfo;  // delay information
	
	int*		IfiringTable;
	int*		EfiringTable;
	int*		randId;
	void*		noiseGenProp;	
} network_ptr_t;

#define MAX_GRP_PER_SNN 100

class CpuSNN
{
	public:

		CpuSNN(int _N, int _M, int _D, char* _name, int mode=CPU_MODE);
		~CpuSNN();

		void dumpHistogram();
		void saveState();
		void saveNetwork();
		void saveFiringInfo();
		void getNode();

		//group creation methods..
		int createGroup(char* _name, int _numN);
		int createGroup(char* _name, int _st, int _end);
		int createInitGroup(char* _name, int _numN, float _a, float _b, float _c, float _d);
		int createInitGroup(char* _name, int _st, int _end, float _a, float _b, float _c, float _d);
		void initGroup(int groupId, float _a, float _b, float _c, float _d);
		void initGroup(int groupId, float _a, float a_sd, float _b, float b_sd, float _c, float c_sd, float _d, float d_sd) ;

		//conection related methods...
		inline void setConnection(int src, int dest, float initWt, uint8_t dVal);
		inline bool checkConnectionOkay(int src, int dest);
		void connect_random(int grpId1, int grpId2, float initWt, int C, uint8_t minDelay,  uint8_t maxDelay);

		// make from each neuron in grpId1 to 'M' neurons in grpId2
		void connect(int grpId1, int grpId2, char* _type, float initWt, int _C, uint8_t minDelay, uint8_t maxDelay);

		// print all the connections...
		void printPostConnection(FILE *fp = stdout);
		void printPreConnection(FILE  *fp = stdout);
		void printConnection(char* fname)
		{
			FILE *fp = fopen(fname, "w");
			printConnection(fp);
			fclose(fp);
		}

		void printConnection(FILE *fp = stdout)
		{
			printPostConnection(fp);
			printPreConnection(fp);
		}

		// print the connection info of grpId
		void printConnection(int grpId, FILE  *fp = stdout)
		{
			printPostConnection(grpId, fp);
			printPreConnection(grpId, fp);
		}
		void printPostConnection(int grpId, FILE  *fp = stdout);
		void printPreConnection(int grpId, FILE  *fp = stdout);

		// run the simulation for n sec
		void runNetwork(int _nsec, int _tstep = 0, int simType = CPU_MODE);

		// input noise related codes...
		void randomNoiseCurrent(float neuronPercentage, float currentStrength, int groupId=-1);

		void printMemoryInfo(FILE* fp=stdout);

		void doSnnSim(int cpu_with_stdp);

		void stdp ( float ALPHA_LTP, float TAU_LTP, float ALPHA_LTD, float TAU_LTD);

		void initThalInput();

		void printSimSummary(FILE *fp = stdout);

		void doECurrentUpdate(int cpu_with_stdp);

		void doICurrentUpdate();

		void globalStateUpdate();

		void updateStateAndFiringTable();

		void snnProperty(char* _name, float val);

		void saveFiringInfo(char* fname, FILE *fpSpikeSave = NULL);

		void dumpHistogram(int* table_sHist, int* table_sdHist, FILE *fph, int sec, char *fname, int histSize=HISTOGRAM_SIZE);

		void printHistogram(char* histName=NULL, bool dontOverWrite=0, unsigned int histSize=HISTOGRAM_SIZE);

		void resetNetwork();

		void resetWeights();

		void setLogCycle(int _cnt, FILE *fp=NULL) {

			//enable or disable logging...
			showLog = (_cnt == 0)? 0 : 1;

			//set the update cycle...
			showLogCycle = _cnt;

			if(fp!=NULL)
				fpProgLog = stdout;

		}

		void spikeMonitor(int gid, float fracMonitor=1.0f, int maxRate=-1);

		void rasterPlot(int gid, char* fname, bool dontOverWrite=false);

		void startCPUTiming();
		void resetCPUTiming();
		void stopCPUTiming();
		void startGPUTiming();
		void resetGPUTiming();
		void stopGPUTiming();

	private:

		void initFiringTables();

		void putHistogram(FILE *fp, int sec, char* fname, unsigned int histSize=HISTOGRAM_SIZE, bool dontOverWrite=false);

		void generateHistogram(int* sHist, int* sdHist, int histSize, FILE *fp);

		void generateHistogram_GPU(int* sHist, int* sdHist, int histSize);

		void updateParameters();

		void reorganizeNetwork();

		int  okaySNNRules();

		void reorganizeDelay();

		void updateRandomProperty();

		/* ..minimize any other wastage in that array by compacting the store */
		void compactConnections();

		/* Here we reorganize all the input connections to the neurons such that
		 the first set is all excitatory connection and next set is inhibitory connections	*/
		void regroupConnections(FILE *fp=NULL);

		/* initialize all the synaptic weights to appropriate values..
		 total size of the synaptic connection is 'length' ...*/
		void initSynapticWeights();

		/* For the given neuron nid, find the group id */
		int findGrpId(int nid);

		/* copy required spikes from the firing buffer to the spike buffer
		for later preprocessing */
		void updateSpikeMonitor();

		// temporary variables created and deleted by network after initialization
		char			*tmpNeuronType;
		int				*tmpPreSynapticIds;
		uint8_t			*tmpSynapticDelay;

	private:
		float prevCpuExecutionTime;
		float cpuExecutionTime;
		float prevGpuExecutionTime;
		float gpuExecutionTime;


		int			currentMode;	// current operating mode
		int			netInitMode;	// initialize operating mode

		// noise generator related info...
		int			numNoise;		// Total number of noise generators
		int			numRandNeurons;	// Total number of neurons selected each time step
		int*		randNeuronId;	// Store the ids of neuron which will get random input current..
		noiseGenProperty_t	noiseGenGroup[MAX_GRP_PER_SNN];

		// properties of the network (number of groups, network name, allocated neurons etc..)
		bool		doneReorganization;

		char		*networkName;
		int			numGrp;
		int			allocatedN;

		// properties of group of neurons size, location, initial weights etc.
		char		*grpName[MAX_GRP_PER_SNN];
		int			grpStartN[MAX_GRP_PER_SNN];
		int			grpEndN[MAX_GRP_PER_SNN];
		int			grpSizeN[MAX_GRP_PER_SNN];
		int			grpNumTraceN[MAX_GRP_PER_SNN];
		float		grpInitWeights[MAX_GRP_PER_SNN];
		char		grpNeuronType[MAX_GRP_PER_SNN];
		char		grpMaxFiringRate[MAX_GRP_PER_SNN];
		int			grpMonitorId[MAX_GRP_PER_SNN];

		// spiking network related info.. neurons, synapses and network parameters
		int			N,M,D,Ne,Ni;
		float       *v, *u, *a, *b, *c, *d, *I;
		short       *Npre;			// stores the number of input connections to the neuron
		short		*Npre_exc;		// stores the number of excitatory input connection to the input
		short       *Npost;			// stores the number of output connections from a neuron.
		int         *lastSpikeTime;	// stores the most recent spike time of the neuron
		float		*wtChange, *wt;		// stores the synaptic weight and weight change of a synaptic connection
		int			*synSpikeTime;	// stores the spike time of each synapse
		int			postSynCnt;
		int			preSynLength;

		int			*cumulativePost;
		int			*cumulativePre;
		int         *postSynapticIds;		// 10 bit syn id, 22 bit neuron id, ordered based on delay
										// allows maximum synapses of 1024 and maximum network
										// size of 4 million neurons, with 64 bit representation
										// we can have larger networks for simulation
		delay_info_t    *postDelayInfo;      // delay information

		FILE*		fpDotty;

		//size of memory used for different parts of the network
		int			neuronInfoSize;
		int			synapticInfoSize;
		int			networkInfoSize;
		int			spikingInfoSize;
		int			debugInfoSize;
		int			addInfoSize;	// includes random number generator etc.

		// firing info
		int		   *EtimeTable;
		int		   *ItimeTable;
		int        *EfiringTable;
		int        *IfiringTable;
		int			maxSpikesInhibitory, maxSpikesExcitatory;

		// stdp setup
		float		ALPHA_LTP, TAU_LTP;
		float		ALPHA_LTD, TAU_LTD;
		bool		STDPenabled;

		//time and timestep
		int			simTimeMs, simTimeSec;					// this is used to store the seconds.
		int			simTime;							// this value is not reset but keeps increasing to its max value.
		int			spikeCountAll1sec, spikeCountInhibitory1sec, spikeCountExcitatory1sec;	// firing counts for each second
		int			spikeCountAll, spikeCountInhibitory, spikeCountExcitatory;	// overall firing counts values

		// properties of the network and synapses..
		float		maxWeight;	// maximum value for the synaptic weights
		float		maxWeightChange;  // maximum value reached by sd

		//cuda keep track of performance...
	    unsigned int timer;
		float		cumExecutionTime;
		float       lastExecutionTime;

		//debug file
		FILE*		fpProgLog;
		FILE*		fpLog;
		int			showLog;
		int			showLogCycle;			// how often do we need to update the log

		//histogram related
		int			stopHistogram;
		int			histFileCnt;
		int			rasterFileCnt;

		//spike monitor code...
		int			numSpikeMonitor;
		int			monGrpId[MAX_GRP_PER_SNN];
		float		monGrpFrac[MAX_GRP_PER_SNN];
		int			monBufferPos[MAX_GRP_PER_SNN];
		int			monBufferSize[MAX_GRP_PER_SNN];
		int*		monBufferFiring[MAX_GRP_PER_SNN];
		int*		monBufferTimeCnt[MAX_GRP_PER_SNN];
		bool		rasterPlotClosed;
		fileInfo_t* rasterFileInfo;
		void		dumpSpikeBuffToFile(FILE *fpSpikeSave, int gid);
		void		updateRasterPlot();
		void		clearRasterData();

		float*		stpu;
		float*		stpx;
		float*		gAMPA;
		float*		gNMDA;
		float*		gGABAa;
		float*		gGABAb;

	public:

		void simMode(int mode)
		{
			if (netInitMode==CPU_MODE && (gpu_Ptrs.allocated == false)) {
				allocate();
			}
			currentMode = mode;
		}

		void saveStateGPUtoCPU()
		{

		}

	private:
		// gpu related info...
		// information about various data allocated at GPU side...
		int			gpu_tStep, gpu_simSec;				// this is used to store the seconds.
		int			gpu_simTime;					// this value is not reset but keeps increasing to its max value.

		network_ptr_t  gpu_Ptrs;
		network_info_t net_Info;
		network_ptr_t  cpu_Ptrs;

		// allocates required memory and then initialize the GPU
		void allocate();

		void copyConnections(network_ptr_t* dest, int kind, int allocateMem);

		void copyState(network_ptr_t* dest, int kind, int allocateMem);

		void copyParameters();

		void initThalInput_GPU(int with_stdp);

		void findFiring_GPU(int with_stdp);

		void updateTimingTable_GPU(int with_stdp);

		void doCurrentUpdate_GPU(int with_stdp);

		void globalStateUpdate_GPU(int with_stdp);

		void doGPUSim(int with_stdp = 1);

		void initGPU();

		void copyFiringInfo_GPU();

		void updateStateAndFiringTable_GPU();

		void showStatus(int simType=CPU_MODE, int onlyCnt = 0);

		void checkInitialization();

};

#endif
